/*
								+---------------------------------+
								|                                 |
								|     ***   BSP Engine   ***      |
								|                                 |
								|  Copyright   -tHE SWINe- 2002  |
								|                                 |
								|            Main.cpp             |
								|                                 |
								+---------------------------------+
*/

#include <windows.h>
#include <malloc.h>
#include <stdio.h>
#include <math.h>

#include "Directx.h"
#include "Matrix.h"
#include "Font_eng.h"
#include "Load_bsp.h"
#include "Raster.h"

#include "Main.h"


#define APPNAME "BSP_Engine"
#define WNDNAME "BSP Engine 1.0"
static int Width = 512;
static int Height = 384;

unsigned __int32 *video;
unsigned __int32 *buffer;

#define PI 3.1415926

static BSP_Tree *svet;
static Matrix cmatrix;

// vytvori soubor bsp.log
// a vlozi do nej uvodni radku
void BSP_Log_Init()
{
	FILE *fw;
	fw = fopen("BSP.log", "w");
	fprintf(fw, "__________BSP - Log__________\n");
	fclose(fw);
}

// pripoji k souboru bsp.log text l_str a odradkuje
void BSP_Log(char *l_str)
{
	FILE *fw;
	fw = fopen("BSP.log", "a");
	fprintf(fw, "%s\n", l_str);
	fclose(fw);
}

// main
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdline, int nCmdShow)
{
	char szAppName[] = APPNAME;
	WNDCLASSEX wndclass;
	HWND hwnd;
	MSG msg;

	BSP_Log_Init();

	wndclass.cbSize			= sizeof (wndclass);
	wndclass.style			= CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc	= Wokno;
	wndclass.cbClsExtra		= 0;
	wndclass.cbWndExtra		= 0;
	wndclass.hInstance		= hInstance;
	wndclass.hIcon			= LoadIcon(hInstance, IDI_APPLICATION);
	wndclass.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground	= (HBRUSH) GetStockObject(BLACK_BRUSH);
	wndclass.lpszMenuName	= 0;
	wndclass.lpszClassName	= szAppName;
	wndclass.hIconSm		= LoadIcon(NULL, IDI_APPLICATION);
	RegisterClassEx(&wndclass);
	hwnd = CreateWindow(szAppName,
		WNDNAME,
		WS_DLGFRAME,
		CW_USEDEFAULT, CW_USEDEFAULT,
		Width, Height,
		NULL,
		NULL,
		hInstance,
		NULL);
	ShowWindow(hwnd, nCmdShow);
	UpdateWindow(hwnd);
	ShowCursor(FALSE);
	// vytvo okno a schov my

	InitDirectX(hwnd, Width, Height);
	// spust directX

	BSP_Log("DDraw is running..");

	Matrix_Init(&cmatrix);
	Matrix_Rotate_Z(&cmatrix, (float)-PI);
	Matrix_Translate(&cmatrix, 0, -40, -40);
	// kamera ok

	BSP_Log("Loading tree ..");

	if((svet = Load_Tree("Svet.BSP")) == NULL)
		return 0;
	Matrix_Init(&svet->matrix);
	// svet ok

	BSP_Log("    OK");

	BSP_Log("Init engine ...");
	
	Init_Engine(Width, Height);
	Init_Font_Engine(Width, Height);
	// engine ok

	BSP_Log("    OK");

	if((buffer = (unsigned __int32*)malloc(Width * Height * 4)) == NULL)
		return 0;

	BSP_Log("Loading textures ...");

	if(!Load_BSP_Textures(svet, "Textures"))
		return 0;
	// nahraje textury

	BSP_Log("    OK");

	BSP_Log("Rendering !");

	SetTimer(hwnd, 0, 20, NULL);

	while(GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	free((void*)buffer);
	// uvoluje pam

	Terminate_Engine();
	Terminate_Font_Engine();
	// uvoluje pam v ostatnch modulech

	TerminateDirectX();
	// ukon directx

	return msg.wParam;
}

// ati technologies filtr quincunx
void Quincunx(unsigned __int32 *src, unsigned __int32 *dest)
{
	int w2, h2, w1;
	int x, y;

	w2 = Width - 1;
	h2 = Height - 1;
	w1 = w2 + 1;

	y = Width * Height;
	for(x = 0; x < y; x ++)
		src[x] = ((src[x] >> 2) & 0x3f3f3f);

	for(y = 0; y < h2; y ++, dest += Width, src += Width) {
		for(x = 0; x < w2; x ++)
			dest[x] = src[x] + src[x + 1] + src[x + w1] + src[x + w2];
		dest[x] = src[x];
	}
	memcpy(dest, src, Width << 2);

	// obrazovy filtr
}

// funkce, zpracovavajici zpravy
// pri sipkach otaci kamerou, ins/del
// hybe kamerou dopredu dozadu
// a kresli obraz
// pri esc. skonci
LRESULT CALLBACK Wokno (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	float speeds[] = {2, 1, 0.5, 3};
	static float a = 0;
	static int j = 0;
	char str[256];

	switch(iMsg) {
	case WM_KEYDOWN:
		if(wParam == 27){
			ShowCursor(TRUE);
			SendMessage(hwnd, WM_CLOSE, 0, 0);
			return 0;
		} else {
			switch (wParam){
			case VK_LEFT:
				Matrix_Rotate_Y(&cmatrix, (float)PI/50);
				break;
			case VK_RIGHT:
				Matrix_Rotate_Y(&cmatrix, (float)-PI/50);
				break;
			case VK_UP:
				Matrix_Rotate_X(&cmatrix, (float)-PI/50);
				break;
			case VK_DOWN:
				Matrix_Rotate_X(&cmatrix, (float)PI/50);
				break;
			case VK_INSERT:
				Matrix_Translate(&cmatrix, 0, 0, -5);
				break;
			case VK_DELETE:
				Matrix_Translate(&cmatrix, 0, 0, 5);
				break;
			default:
				break;
			}
		} // posouvani kamery sipkama
		return 0;

	case WM_TIMER:
		memset(buffer, 0, 4 * Width * Height);
		// vycisti obrazovku

		a += (float)0.1;
		if(a > 6.28)
			a = 0;

		TraverseTree(&cmatrix, svet, buffer);
		// kresli svet

		if(InitVideoBuffer(&video)) {
			///
			memcpy(video, buffer, 4 * Width * Height);
			//Quincunx(buffer, video);
			/// zkopruje obraz

			if(j % 10 == 1)
				sprintf(str, "%d.st Frame", j ++);
			else if(j % 10 == 2)
				sprintf(str, "%d.nd Frame", j ++);
			else
				sprintf(str, "%d.th Frame", j ++);
			String_Out_NoMask(str, 10, 10, video);
			// vypise cislo snimku
		
			FlipVideoBuffer();
		}
		return 0;

	case WM_DESTROY:
		PostQuitMessage (0);
		return 0;
	}

	return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}





