/*
                                +---------------------------------+
                                |                                 |
                                |     ***   Rasterizer   ***      |
                                |                                 |
                                |  Copyright   -tHE SWINe- 2002  |
                                |                                 |
                                |           Raster.cpp            |
                                |                                 |
                                +---------------------------------+
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "Load_BSP.h"
#include "Matrix.h"
#include "Pyramid.h"
#include "Font_Eng.h"

#include "Raster.h"

#define PI 3.14159265358979323846263279502884

struct Poly_Edge {                            
    int x, y, num, dxdy;            
};

static int n_Width, n_Height;
// vyska a sirka obrazovky

static float z_delta;
static Pyramid pyramid;
// klipovaci pyramida

static Poly_Edge left_edge_buff[7];
static Poly_Edge right_edge_buff[7];
static Poly_Edge *p_left_edge;
static Poly_Edge *p_right_edge;
static int n_left_edge_num, n_right_edge_num;
// edge

static float dzdx, dzdy, begz;
static float dudx, dudy, begu;
static float dvdx, dvdy, begv;
// interpolace

static int A_1, A_2, B_1;
static unsigned __int32 *_texture;
// texturovani

static Matrix m, im;
// transforman matice

int Init_Engine(int n_New_Width, int n_New_Height)
{
    n_Width = n_New_Width;
    n_Height = n_New_Height;
    // nastav velikost obrazu

    z_delta = (float)fabs((float)n_Width / ((float)tan(90 * (PI / 180) / 2) * 2));
    // perspektivni zkresleni pro hel 90

    SetFOV((float)n_Width - 4, (float)n_Height - 4, z_delta, &pyramid);
    // klipovaci pyramida (je o trochu men kvli nepesnostem)

    return 1;
}

void Terminate_Engine()
{
    // nic --
}

int Add_Edge(Vertex *v_cur, Vertex *v_next, Poly_Edge *edge)
{
    float y1, y2;
    float dxdy;
    int num;

    y1 = (float)ceil(v_cur->y);
    y2 = (float)ceil(v_next->y);
    num = (int)(y2 - y1);

    if(!num)
        return 0;

    dxdy = (v_next->x - v_cur->x) / (v_next->y - v_cur->y);
    edge->dxdy = (int)(dxdy * 65536);
    edge->x = (int)((v_cur->x + (y1 - v_cur->y) * dxdy) * 65536);
    edge->y = (int)y1;
    edge->num = num;
    // spot prvn x a y, sklon (dxdy) a poet obrazovch dek

    return 1;
}

int Create_Edges(Polygon *p_poly)
{
    int n_bottom, n_top, n_cur;
    float f_min_y;
    float f_max_y;

    n_left_edge_num = 0;
    n_right_edge_num = 0;
    // poty hran

    f_min_y =  0x10000;
    f_max_y = -0x10000;
    n_cur = 0;
    while(n_cur < p_poly->n_vertex_num) {
        if(p_poly->vertex[n_cur].y < f_min_y) {
            f_min_y = p_poly->vertex[n_cur].y;
            n_top = n_cur;
        }
        if(p_poly->vertex[n_cur].y > f_max_y) {
            f_max_y = p_poly->vertex[n_cur].y;
            n_bottom = n_cur;
        }
        n_cur ++;
    }
    // najde nejvrchnj a nejspodnj vrchol

    if((int)ceil(f_min_y) >= (int)ceil(f_max_y))
        return 0;
    // zkontroluje zda se vejde do rastru

    n_cur = n_top;
    while(n_cur != n_bottom) {
        if(Add_Edge(&p_poly->vertex[n_cur], &p_poly->vertex[(n_cur + 1) %
           p_poly->n_vertex_num], &right_edge_buff[n_right_edge_num]))
            n_right_edge_num ++;
        n_cur = (n_cur + 1) % p_poly->n_vertex_num;
    }
    // sjede pravou vtev

    n_cur = n_bottom;
    while(n_cur != n_top) {
        if(Add_Edge(&p_poly->vertex[(n_cur + 1) % p_poly->n_vertex_num],
           &p_poly->vertex[n_cur], &left_edge_buff[n_left_edge_num]))
            n_left_edge_num ++;
        n_cur = (n_cur + 1) % p_poly->n_vertex_num;
    }

    p_left_edge = left_edge_buff + n_left_edge_num - 1;
    p_right_edge = right_edge_buff;

    return 1;
}

int Compute_Interpolation_Consts(Polygon *p_poly)
{
    float x[2], y[2], z[2], u[2], v[2];
    float a, b, c;
    
    x[0] = p_poly->vertex[1].x - p_poly->vertex[0].x;
    x[1] = p_poly->vertex[2].x - p_poly->vertex[0].x;

    y[0] = p_poly->vertex[1].y - p_poly->vertex[0].y;
    y[1] = p_poly->vertex[2].y - p_poly->vertex[0].y;
    
    z[0] = (1 / p_poly->vertex[1].z) - (1 / p_poly->vertex[0].z); 
    z[1] = (1 / p_poly->vertex[2].z) - (1 / p_poly->vertex[0].z);
    
    u[0] = 1 / p_poly->vertex[1].z * p_poly->vertex[1].u -
        1 / p_poly->vertex[0].z * p_poly->vertex[0].u;
    u[1] = 1 / p_poly->vertex[2].z * p_poly->vertex[2].u -
        1 / p_poly->vertex[0].z * p_poly->vertex[0].u;

    v[0] = 1 / p_poly->vertex[1].z * p_poly->vertex[1].v -
        1 / p_poly->vertex[0].z * p_poly->vertex[0].v;
    v[1] = 1 / p_poly->vertex[2].z * p_poly->vertex[2].v -
        1 / p_poly->vertex[0].z * p_poly->vertex[0].v;
    // vektory (od vrocholu 0 do 1, 2 pro kadou souradnici)
    // vektory u a v jsou jet nsobene 1/z

    a = y[0] * z[1] - y[1] * z[0];
    b = z[0] * x[1] - z[1] * x[0];
    c = x[0] * y[1] - x[1] * y[0];
    // cross product

    if(c == 0) 
        return 0;
    
    dzdx = - a / c;
    dzdy = - b / c;
    // uren hodnot podle rovnice

    begz = (1 / p_poly->vertex[0].z) - p_poly->vertex[0].x * dzdx - p_poly->vertex[0].y * dzdy;
    // rychlej uren - d / c
    // Z

    a = y[0] * u[1] - y[1] * u[0];
    b = u[0] * x[1] - u[1] * x[0];
    //c = x[0] * y[1] - x[1] * y[0];
    // cko u bude pod stejn

    dudx = - a / c;
    dudy = - b / c;
    // uren hodnot podle rovnice

    begu = (1 / p_poly->vertex[0].z * p_poly->vertex[0].u) -
        p_poly->vertex[0].x * dudx - p_poly->vertex[0].y * dudy;
    // rychlej uren - d / c
    // U
    
    a = y[0] * v[1] - y[1] * v[0];
    b = v[0] * x[1] - v[1] * x[0];
    //c = x[0] * y[1] - x[1] * y[0];
    // cko u bude pod stejn

    dvdx = - a / c;
    dvdy = - b / c;
    // uren hodnot podle rovnice

    begv = (1 / p_poly->vertex[0].z * p_poly->vertex[0].v) -
        p_poly->vertex[0].x * dvdx - p_poly->vertex[0].y * dvdy;
    // rychlej uren - d / c
    // V

    return 1;
}

void Interpolate_Segment(int u1, int v1, int du, int dv,
	unsigned __int32 *video, unsigned __int32 *v2)
{
    do {
        *video = _texture[((u1 >> 16) & A_1) + ((v1 >> B_1) & A_2)];
        // nan texturu

        video ++;
        // nov pozice

        u1 += du;
        v1 += dv;
        // interpolace souadnic
    } while(video < v2);
}

void Draw_Segment(int c, int l, float u,
	float v, float z, unsigned __int32 *video)
{
    int du, u1, u2;
    int dv, v1, v2;
    float z1;

    z1 = (0x10000 / z);
    // pevod zptky na linern, do 16:16 fp

    u1 = (int)(u * z1);
    v1 = (int)(v * z1);
    // u a v se pevede troku jinak

    while(c >= 16) {
        z += dzdx * 16;
        u += dudx * 16;
        v += dvdx * 16;
        // pro optimalizaci pepotme jen kazd 16. pixel

        z1 = (0x10000 / z);
        // uren daliho linernho z
        
        u2 = (int)(u * z1);
        v2 = (int)(v * z1);
        // u a v se pevede troku jinak

        du = (u2 - u1) / 16;
        dv = (v2 - v1) / 16;
        // spot delty

        Interpolate_Segment(u1, v1, du, dv, video, video + 16);
        // kreslc smyka

        c -= 16;
        // sn poet pixel o 16

        video += 16;
        // posune se dl

        u1 = u2;
        v1 = v2;
        // vyuije pedpotan hodnoty
    }

    if(c > 0) { // kresl zbytek lajny (pokud dlka nen nsobkem 16)
        z += dzdx * c;
        u += dudx * c;
        v += dvdx * c;

        z1 = (0x10000 / z);
        u2 = (int)(u * z1);
        v2 = (int)(v * z1);

        du = (u2 - u1) / c;
        dv = (v2 - v1) / c;

        Interpolate_Segment(u1, v1, du, dv, video, video + c);
    }
}

int DrawPolygon(Polygon *p, unsigned __int32 *video)
{
    int start_y, left_x, right_x;
    float z, u, v;
    int l, c;

    if(!Create_Edges(p)) 
        return 0;

    if(!Compute_Interpolation_Consts(p)) 
        return 0;

    start_y = p_right_edge->y;
    left_x = p_left_edge->x;
    right_x = p_right_edge->x;
    // zkopruje vlastnosti edge

    video += n_Width * start_y;
    // nastav video na prvn dku rastru

    z = begz + dzdy * start_y;
    u = begu + dudy * start_y;
    v = begv + dvdy * start_y;
    // z prvniho radku rastru

    while(1) {
        c = left_x >> 16;
        l = right_x >> 16;
        // pevod zptky na int

        c -= l ++;

        if(c > 0)
            Draw_Segment(c, l, u + dudx * l, v + dvdx * l, z + dzdx * l, video + l);
        // kresl vodorovny segment

        video += n_Width;
        z += dzdy;
        u += dudy;
        v += dvdy;
        // posun na dalsi ek obrazovky

        if(-- p_left_edge->num == 0) {
            if(p_left_edge -- == left_edge_buff)
                return 1;
            // pokud dojde na konec bufferu, je konec polygonu
            left_x = p_left_edge->x;
        } else 
            left_x += p_left_edge->dxdy;

        if(-- p_right_edge->num == 0) {
            p_right_edge ++;
            right_x = p_right_edge->x;
        } else 
            right_x += p_right_edge->dxdy;
    }

    return 1;
}

void TransformDraw_Face(Face *p_face,
	unsigned __int32 *video, Texture *p_texture)
{
    float tx, ty, tz;
    float x, y, z;    
    Polygon poly;
    Vertex v[3];
    Face face;
    int i;

    face.vertex[0] = &v[0];
    face.vertex[1] = &v[1];
    face.vertex[2] = &v[2];
    // nastav vertexy pro nai doasnou face

    if((p_face->normal.c * im[3][2] +
       p_face->normal.a * im[3][0] +
       p_face->normal.b * im[3][1] +
       p_face->normal.d) > 0.1)
        return;
    // backface culling

    for(i = 0; i < 3; i ++){
        x = p_face->vertex[i]->x;
        y = p_face->vertex[i]->y;
        z = p_face->vertex[i]->z;
        //
        tx = x * m[0][0] + y * m[1][0] + z * m[2][0] + m[3][0];
        ty = x * m[0][1] + y * m[1][1] + z * m[2][1] + m[3][1];
        tz = x * m[0][2] + y * m[1][2] + z * m[2][2] + m[3][2];
        //
        face.vertex[i]->x = tx;
        face.vertex[i]->y = ty;
        face.vertex[i]->z = tz;
        //
        face.vertex[i]->u = p_face->vertex[i]->u * p_texture->n_Width;
        face.vertex[i]->v = p_face->vertex[i]->v * p_texture->n_Height;
    }

    poly = Clip_Face(&face, &pyramid);
    // proene face pyramidou

    if(poly.n_vertex_num) { // pokud je face vidt, bude to nenulov slo
        for(i = 0; i < poly.n_vertex_num; i ++) {
            if(poly.vertex[i].z < 0.1) {
                poly.n_vertex_num = 0;
                break;
            }
            // pokud je z - souadnice moc mal, zahodme polygon

            poly.vertex[i].x = (n_Width / 2) + (poly.vertex[i].x *
                z_delta) / poly.vertex[i].z;
            poly.vertex[i].y = (n_Height / 2) + (poly.vertex[i].y *
                z_delta) / poly.vertex[i].z;
            // perspektivn korekce
        }

        _texture = p_texture->texture;
        A_1 = p_texture->A_1;
        A_2 = p_texture->A_2;
        B_1 = p_texture->B_1;
        // nastav vlastnosti textury

        DrawPolygon(&poly, video);
        // vykresl polygon na obrazovku
    }
}

// prochz strom a kresl polygony odzadu dopedu
void _fastcall BSP_Traversal(BSP_Node *node, unsigned __int32 *vram, BSP_Tree *tree)
{
	float pl_eq;
	int i;

	pl_eq = node->rootpl.d + node->rootpl.a * im[3][0]
		  + node->rootpl.b * im[3][1]
		  + node->rootpl.c * im[3][2];
	// spot rovnici roviny pro root

	// root je ped kamerou
	if(pl_eq > 0.01) {
		if(node->flags & _Have_Back_Node)
			BSP_Traversal(node->back, vram, tree);
		// pokud m zadn nodu, projde ji ...

		for(i = 0; i < node->n_onplane_num; i ++) {
			TransformDraw_Face(&node->onplane[i], vram,
				tree->material[node->onplane[i].n_material].texture);
		}
		// kresl fejsy

		if(node->flags & _Have_Front_Node)
			BSP_Traversal(node->front, vram, tree);
		// pokud m pedn nodu, projde ji ...
	} else if(pl_eq < -0.01) {
		// root je za kamerou
		if(node->flags & _Have_Front_Node)
			BSP_Traversal(node->front, vram, tree);
		// pokud m pedn nodu, projde ji ...

		for(i = 0; i < node->n_onplane_num; i ++) {
			TransformDraw_Face(&node->onplane[i], vram,
			    tree->material[node->onplane[i].n_material].texture);
		}
		// kresl fejsy

		if(node->flags & _Have_Back_Node)
			BSP_Traversal(node->back, vram, tree);
		// pokud m zadn nodu, projde ji ...
	} else {
		// root je v rovin kamery (nen vidt)
		if(node->flags & _Have_Back_Node)
			BSP_Traversal(node->back, vram, tree);
		// pokud m zadn nodu, projde ji ...

		// - nekresl nic, kdy nen vidt

		if(node->flags & _Have_Front_Node)
			BSP_Traversal(node->front, vram, tree);
		// pokud m pedn nodu, projde ji ...
	}
}

// kresl bsp-strom tree z pohledu kamery do video
void TraverseTree(Matrix *m_camera, BSP_Tree *tree, 
	unsigned __int32 *video)
{
	Matrix_Inverse(m_camera, &m);
    Matrix_Multiply(&m, &tree->matrix, &m);
    Matrix_Inverse(&m, &im);
    // spot transforman matice

	BSP_Traversal(tree->root, video, tree);
	// zavola prochazeni stromu
}

Texture *Load_Texture(char *filename)
{
    Texture *temp;
    BMP *bmp;

    if((bmp = Load_TrueColorBMP(filename)) == NULL)
        return NULL;
    // nahraje bitmapu

    if((temp = (Texture*) malloc(sizeof(Texture))) == NULL) {
        free((void*)bmp->bytes);
        free((void*)bmp);
        // pokud neuspeje, uvolni bitmapu
        return NULL;
    }
    // naalokuje texturu

    temp->n_Width = bmp->n_Width;
    temp->n_Height = bmp->n_Height;
    temp->texture = bmp->bytes;
    // zkopiruje data

    free((void*)bmp);
    // uvolni jiz nepotrebnou bitmapu

    temp->A_1 = temp->n_Width - 1;
    temp->A_2 = temp->n_Width * (temp->n_Height - 1);
    temp->B_1 = 16 - (int)(log((float)temp->n_Height) / log(2) + 1e-7);
    // A_1 je maximalni x-ova souradnice
    // A_2 je maximalni y-ova souradnice (nasobena sirkou bitmapy)
    // B_1 je 16 - mocnina dvou vysky textury (Height = 256 <=> B_1 = 8)

    return temp;
}

// nahraje textury podle jmen materil v bsp-stromu
// (textury jsou hledny ve sloce dir)
int Load_BSP_Textures(BSP_Tree *tree, char *dir)
{
	char path[256];
	int i;

	if(dir[strlen(dir) - 1] == '\\')
		dir[strlen(dir) - 1] = 0;
	// pokud adres obsahuje lomtko, umae se
	// (je to jednodu ne ho pidat ;-) )

	for(i = 0; i < tree->n_material_num; i ++) {
		sprintf(path, "%s\\%s.bmp", dir, tree->material[i].name);
		// vytvo etzec s adresou k bitmap, reprezentujc texturu

		tree->material[i].texture = Load_Texture(path);
		// nahraje texturu

		if(tree->material[i].texture == NULL)
			return 0;
		// soubor neexistuje..
	}

	return 1;
}